
import processing.core.PApplet;
import processing.serial.Serial;
import xbee.XBeeDataFrame;
import xbee.XBeeReader;

// Interface for controlling the Arduino chip
class ArduinoControl {
  
  PApplet parent;
  
  public void Setup ( PApplet theApplet ) {
    Connect( theApplet );
    parent = theApplet;
  }
  
  public void Connect ( PApplet theApplet ) {
    //Iterate over each available serial port
    serial = null;
    String[] portNames = Serial.list();
    for( String portName : portNames ) { //Ping each port until we get a response
    
      if( !portName.contains( "cu.usbserial" ) && serial == null ) {
        try {
          System.out.print( "Pinging " + portName + "\n" );
          serial = new Serial( theApplet, portName, 57600 );
          //Thread.sleep( 6000 );
          serial.clear();
          serial.write( "+++" );
          Thread.sleep( 1100 );
          
          if( serial.available() >= 2 ) { //Check the response
            String response = serial.readString();
            System.out.print( "Received \"" + response + "\"\n" );
            
            // If we received "OK", it's an XBee
            if( response.charAt( 0 ) == 'O' ) {
              serial.write( "ATCN\n" );
              xbee = new XBeeReader( theApplet, serial );
              xbee.startXBee();
              break;
            }
            // If we received "KK", it's a Serial connection
            else if( response.charAt( 0 ) == 'K' ) {
              break;
            }
          }
          serial = null;
        } catch( Exception e ) { //If any port fails, print the exception and continue
          System.out.print( "Failed to connect to " + portName + "\n" );
          serial = null;
        }
      }
    }
  }
  
  public void Read () {
    if( xbee != null ) { //Read from XBee
      XBeeDataFrame data = xbee.getXBeeReading();
      if( data == null ) return;
      
      if( data.getApiID() == xbee.SERIES1_RX16PACKET ) {
        int addr = data.getAddress16();
        int[] bytes = data.getBytes();
        
        System.out.print( parent.millis() - lastReadMs + "ms [" + addr + "]:" );
        System.out.println( DataToString( bytes ) );
        
        lastReadMs = parent.millis();
      }
      
      xbee.dataFrame = null;
    } else if( serial != null ) { //Read from Serial
      if( serial.available() > 0 ) {
        String response = serial.readString();
        
        System.out.print( parent.millis() - lastReadMs + "ms:" );
        System.out.println( response );
        
        lastReadMs = parent.millis();
      }
    }
  }
  
  void SetDestAddress ( int destAddr ) {
    // Interpret 0 as broadcast, since we don't need to index the coordinator
    if( destAddr == 0 ) destAddress = 0xFFFF;
    else destAddress = destAddr;
  }
  
  void ClearDevices () {
    sendData( new int[] { 0 } );
  }
  
  void sendData ( int[] data ) {
    if( xbee != null ) {
      xbee.sendDataString16( destAddress, data );
    } else if( serial != null ) {
      serial.write( (byte)data.length );
      for( int i = 0; i < data.length; ++i )
        serial.write( (byte)data[i] );
    }
  }
  
  void xbeeDelay () {
    if( xbee != null ) parent.delay( 66 );
  } //Give the data time to send #HACKHACK#
  
  void AddXYServo ( int xPin, int yPin, int zPin ) {
    sendData( new int[] { 1, xPin, yPin, zPin } );
    xbeeDelay();
  }
  
  void AddPWM3 ( int xPin, int yPin, int zPin ) {
    sendData( new int[] { 2, xPin, yPin, zPin } );
    xbeeDelay();
  }
  
  void AddSwitchArray ( int clockPin, int dataPin, int resetPin ) {
    sendData( new int[] { 3, clockPin, dataPin, resetPin } );
    xbeeDelay();
  }
  
  void AddScreen () {
    sendData( new int[] { 4 } );
    xbeeDelay();
  }
  
  void AddCmdBuffer ( int addAsDevice ) {
    sendData( new int[] { 5, addAsDevice } );
    xbeeDelay();
  }
  
  void SaveCmdBuffer () {
    sendData( new int[] { 53 } );
    xbeeDelay();
  }
  
  void LoadCmdBuffer () {
    sendData( new int[] { 54 } );
    xbeeDelay();
  }
  
  void ExecuteCmdBuffer () {
    sendData( new int[] { 55 } );
    xbeeDelay();
  }
  
  void ClearCmdBuffer () {
    sendData( new int[] { 56 } );
    xbeeDelay();
  }
  
  void SendPosition ( int x, int y, int deviceIndex ) {
    sendData( new int[] { 50, deviceIndex, x, y } );
  }
  
  void SendPosition ( int x, int y, int z, int deviceIndex ) {
    sendData( new int[] { 51, deviceIndex, x, y, z } );
  }
  
  void SendSwitchArray ( char[] bytes, int deviceIndex ) {
//    if( bytes.length < 4 ) return;
    sendData( new int[] { 52, deviceIndex, bytes[0], bytes[1], bytes[2], bytes[3] } );
  }
  
  void EchoDevices () {
    sendData( new int[] { 100 } );
  }
  
  private String DataToString ( int[] data ) {
    String str = new String();
    for( int curValue : data ) {
      if( curValue != 0 ) str += (char)curValue;
    }
    return str;
  }
  
  // Private members
  private int        destAddress = 0xFFFF;
  private Serial     serial;
  private XBeeReader xbee;
  private int        lastReadMs  = 0;
}
