/*
  Caffeine Processing Libraries
  
  (c) copyright
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
 */
 
package com.fieldfx.ui;

import com.fieldfx.math.LineSeg2;
import com.fieldfx.math.Vector2;
 
// -------------------------------------------------------------------------------------------------------------- //
/** UIBox provides all the implementation for a movable, rotatable, scalable and resizable box. It is the first 
  * instantiable element type in the Caffeine user interface library. The box has a default appearence, can be 
  * created by {@link UIFactory} and it will be visible when {@link UIManager.draw}() is called. <br />
  * <br />
  * It serves as a base class for the following elements:
  <ul>
    <li>{@link UIFrame}</li>
    <li>{@link UIKnob}</li>
    <li>{@link UIButton}</li>
    <li>{@link UISlider}</li>
    <li>{@link UITextbox}</li>
  </ul>
  * @author Jack Kern
  */
public class UIBox extends UIElement implements UITransform
{
  // -------------------------------------------------------------------------------------------------------------- //
  // Drag Location Indices
  @SuppressWarnings("unused")
  private final static int DRAG_NONE         = 0;
  private final static int DRAG_TOPLEFT      = 1;
  private final static int DRAG_TOPRIGHT     = 2;
  private final static int DRAG_BOTTOMLEFT   = 3;
  private final static int DRAG_BOTTOMRIGHT  = 4;
  private final static int DRAG_TOP          = 5;
  private final static int DRAG_RIGHT        = 6;
  private final static int DRAG_BOTTOM       = 7;
  private final static int DRAG_LEFT         = 8; 
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** The size of the element. */
  protected   Vector2     size             = new Vector2();
  /** The minimum allowed size of the element. */
  protected   Vector2     minSize          = new Vector2();
  /** The maximum allowed size of the element. */
  protected   Vector2     maxSize          = new Vector2();
  /** The element's current velocity. */
  protected   Vector2     velocity         = new Vector2();
  /** The element's current rotation in radians. */
  protected   float       rotation         = 0.0f;
  /** The element's current scale */
  protected   float       scale            = 1.0f;
  /** How fast this element's momentum will decay */
  protected   float       momentumDecay    = 0.85f;
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** Have we registered a mouse drag on the top edge? */
  private     boolean     hitTop           = false;
  /** Have we registered a mouse drag on the bottom edge? */
  private     boolean     hitBottom        = false;
  /** Have we registered a mouse drag on the left edge? */
  private     boolean     hitLeft     = false;
  /** Have we registered a mouse drag on the right edge? */
  private     boolean     hitRight    = false;
  /** The current closest point on the top egde to the mouse 
      cursor in our parent container's space. */
  private     Vector2     closestTop;
  /** The current closest point on the right egde to the mouse 
      cursor in our parent container's space.   */
  private     Vector2     closestRight;
  /** The current closest point on the bottom egde to the mouse 
      cursor in our parent container's space.   */
  private     Vector2     closestBottom;
  /** The current closest point on the bottom egde to the mouse 
      cursor in our parent container's space.   */
  private     Vector2     closestLeft;

  // -------------------------------------------------------------------------------------------------------------- //
  /** This class is a container for the {@link UIBox}'s properties. */
  public class UIBoxProperties
  {
    /** Does this box have momentum enabled? */
    public boolean              momentumEnabled  = false;
    /** Does this box have rotation by mouse enabled? */
    public boolean              rotation         = false;
    /** Does this box have resize by mouse enabled? */
    public boolean              resize           = false;
    /** Does this box have scale by mouse enabled? */
    public boolean              scale            = false;
    /** Is this box draggable by mouse? */
    public boolean              draggable        = true;
    /** Is this box visible? */
    public boolean              visible          = true;
    /** Does this box have a shadow? */
    public boolean              shadow           = true;
  }
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** This box's behavior properties. */
  protected UIBoxProperties     props            = new UIBoxProperties();
  /** This box's parent transform. */
  protected UITransform         parentTransform  = null;

  
  // -------------------------------------------------------------------------------------------------------------- //
  /** {@inheritDoc} */
  protected UIBox ( UIManager parent, UIContainer container ) {
    super( parent, container );
    
    // Connect the transforms
    setParentTransform( container.getView() );

    onResize();
  }

  // -------------------------------------------------------------------------------------------------------------- //
  //public    void         enable              ( int     flags     )  {     }
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** Enables and disables rotation of this element. If rotation is enabled, users will be able 
   *  to grab the element by its edges and rotate it with the mouse. Defaults to false.<br />
   *  <br />
   *  Note - Rotation works best on it's own or with scale enabled, but not resize.<br />
   *  @param   rotation  true to enable false to disable
   */  
  public    void         enableRotation      ( boolean rotation  )         { this.props.rotation  = rotation;  }
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** Enables and disables resizing of this element. If resizing is enabled, users will be 
   *  able to resize the element by grabbing and draging any edge. This is different than
   *  scale in that the internal size changes, so the aspect ratio can change and child elements 
   *  stay the same size -- there is just more or less room available. Defaults to false. <br />
   *  <br />
   *  Note - Resize works best when rotation and scale are disabled. <br />
   *  @param   resize    true to enable false to disable
   */    
  public    void         enableResize        ( boolean resize    )         { this.props.resize    = resize;    }

  // -------------------------------------------------------------------------------------------------------------- //
  /** Enables and disables scaling of this element. If scale is enabled, users will be
   *  able to scale the element by grabbing and dragging any edge. It stays the same aspect
   *  ratio, and child elements will become larger or smaller on the screen. Defaults to false. <br />
   *  <br />
   *  Note - Scale works best by itself or with rotation enabled, but not with resize enabled. <br />
   *  @param   scale  true to enable false to disable
   */      
  public    void         enableScale         ( boolean scale     )         { this.props.scale     = scale;     }
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** Enables and disables dragging of this element. If drag is enabled, users will be able
   *  to grab the element by clicking inside its client area and dragging the mouse. Defaults to false. <br />
   *  <br />
   *  @param draggable   true to enable false to disable */
  public    void         enableDrag          ( boolean draggable )         { this.props.draggable = draggable; }
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** Enables and disables drawing of the shadow for this element. Defaults to true. <br />
   *  <br />
   *  @param shadow      true to enable false to disable
   */
  public    void         enableShadow        ( boolean shadow    )         { this.props.shadow    = shadow;    }
  
  // -------------------------------------------------------------------------------------------------------------- //
  /** Sets the position of the UIBox in  coordinates
   *  @param  x
   *  @param  y
   *  @return UIBox
   */
  public    UIBox        setPos              ( float x, float y )          { return (UIBox)super.setPos(x,y); }
  public    UIBox        setPos              ( Vector2 pos )               { return (UIBox)super.setPos(pos); }  
  
  // -------------------------------------------------------------------------------------------------------------- //
  public    UIBox        setHeight           ( float height )              { size.y = height;               validateSize(); onResize(); return this; }
  public    UIBox        setWidth            ( float width )               { size.x = width;                validateSize(); onResize(); return this; }  
  public    UIBox        setSize             ( float width, float height ) { size.set( width, height );     validateSize(); onResize(); return this; }   
  public    UIBox        setSize             ( Vector2 size )              { this.size.set( size );         validateSize(); onResize(); return this; }  
  public    UIBox        setMinSize          ( float width, float height ) { minSize.set( width, height );  validateSize(); onResize(); return this; }     
  public    UIBox        setMinSize          ( Vector2 size )              { return setMinSize( size.x, size.y ); }
  
  public    UIBox        setRot              ( float rotation )            { this.rotation = rotation;  return this; }
  public    UIBox        setScale            ( float scale )               { this.scale    = scale;     return this; }  

  
  public    UIBox        show                ( )                    { return (UIBox)super.show(); }
  public    UIBox        hide                ( )                    { return (UIBox)super.hide(); }  

  public    UIBox        setFocus            ( )                    { return (UIBox)super.setFocus();       }   
  public    UIBox        setFocus            ( boolean bSet )       { return (UIBox)super.setFocus( bSet ); }     
  
  public    UIBox        bringToFront        ( )                    { return (UIBox)super.bringToFront();   }
  public    UIBox        sendToBack          ( )                    { return (UIBox)super.sendToBack();     }

  // -------------------------------------------------------------------------------------------------------------- //
  public    float        getHeight           ( )                    { return size.y;   }
  public    float        getWidth            ( )                    { return size.x;   } 
  public    Vector2      getSize             ( )                    { return size;     }
  public    Vector2      getMinSize          ( )                    { return minSize;  }
  public    float        getRot              ( )                    { return rotation; }
  public    float        getScale            ( )                    { return scale;    }
  
  public    void         onScale             ( )                    { }
  public    void         onRotate            ( )                    { }
  public    void         onResize            ( )                    { }
  
  
  // -------------------------------------------------------------------------------------------------------------- //
  public    void         update              ( )                    { updateMomentum(); }

  // -------------------------------------------------------------------------------------------------------------- //
  public    Vector2      getClientMouse      ( )                    { return clientToWorld( new Vector2( getApp().mouseX, getApp().mouseY ) ); }
  
  // -------------------------------------------------------------------------------------------------------------- //
  public    Vector2      getClientMouseOnly  ( )
  {
    Vector2 clientMouse = getClientMouse();
    if( clientMouse.x > 0 && clientMouse.x < size.x &&
        clientMouse.y > 0 && clientMouse.y < size.y )
    {
      return clientMouse;
    }
    return null;
  }
  // -------------------------------------------------------------------------------------------------------------- //
  private void validateSize()
  {
    if( size.x < minSize.x ) size.x = minSize.x;
    if( size.y < minSize.y ) size.y = minSize.y;
  }  
  
  // -------------------------------------------------------------------------------------------------------------- //
  public void transform( UIDrawer d ) { transform( d, false ); }   
  public void transform( UIDrawer d, boolean bVector )
  { 
    if( !bVector )
      d.translate (  getPos().x, getPos().y );    

    d.translate ( getSize().x * 0.5f, getSize().y * 0.5f );
    d.rotate    ( rotation );
    d.scale     ( scale    );
    d.translate ( getSize().x * -0.5f, getSize().y * -0.5f );
  }
  // -------------------------------------------------------------------------------------------------------------- //
  public void transformInv( UIDrawer d ) { transformInv( d, false ); }   
  public void transformInv( UIDrawer d, boolean bVector )
  { 
    d.translate (  getSize().x * 0.5f, getSize().y * 0.5f );
    d.scale     (  1.f / scale );
    d.rotate    ( -rotation  );
    d.translate (  getSize().x * -0.5f, getSize().y * -0.5f  );
    
    if( !bVector )
      d.translate ( -getPos().x, -getPos().y );    
  }

  // -------------------------------------------------------------------------------------------------------------- //
  private void worldToClientImpl( Vector2 out, Vector2 client, byte flags )
  {
    out.set ( client.x, client.y );

    out.dec ( getSize().x * 0.5f, getSize().y * 0.5f );
    
    if( ( flags & UITRANS_SCALE ) > 0 )
      out.muleq( getScale() );
      
    if( ( flags & UITRANS_ROTATE ) > 0 )
      out.roteq( getRot() );

    out.inc( getSize().x * 0.5f, getSize().y * 0.5f );  
    
    if( ( flags & UITRANS_TRANSLATE ) > 0 )
      out.inc( getPos().x, getPos().y );

  }
  // -------------------------------------------------------------------------------------------------------------- //
  private void clientToWorldImpl( Vector2 out, Vector2 world, byte flags )
  {
    out.set( world.x, world.y );
  
    if( ( flags & UITRANS_TRANSLATE ) > 0 )
      out.dec( getPos().x, getPos().y );

    out.dec( getSize().x * 0.5f, getSize().y * 0.5f );
    
    if( ( flags & UITRANS_ROTATE ) > 0 )
      out.roteq( -getRot() );
    
    if( ( flags & UITRANS_SCALE ) > 0 )
      out.muleq ( 1.f / getScale() );

    out.inc( getSize().x * 0.5f, getSize().y * 0.5f );
  }

  // -------------------------------------------------------------------------------------------------------------- //
  public  Vector2  clientToWorld( Vector2 source                                       ) { return this.clientToWorld( new Vector2(), source );              }
  public  Vector2  clientToWorld( Vector2 source,                int depth             ) { return this.clientToWorld( new Vector2(), source, depth );       }
  public  Vector2  clientToWorld( Vector2 source,                           byte flags ) { return this.clientToWorld( new Vector2(), source, flags );       }
  public  Vector2  clientToWorld( Vector2 source,                int depth, byte flags ) { return this.clientToWorld( new Vector2(), source, depth, flags ); }
  public  Vector2  clientToWorld( Vector2 dest, Vector2 source                         ) { return this.clientToWorld( dest, source,          UITRANS_ALL ); }
  public  Vector2  clientToWorld( Vector2 dest, Vector2 source, int depth              ) { return this.clientToWorld( dest, source,  depth,  UITRANS_ALL ); }
  public  Vector2  clientToWorld( Vector2 dest, Vector2 source,             byte flags ) { return this.clientToWorld( dest, source,   9999,  flags );       }
  public  Vector2  clientToWorld( Vector2 dest, Vector2 source, int depth,  byte flags )
  {
    if( depth < 0 ) return source;
    if( parentTransform == null ) this.clientToWorldImpl( dest, source, flags );
    else
    {
      parentTransform.clientToWorld( dest, source, depth-1, flags );
      this.clientToWorldImpl( dest, dest.get(), flags );
    }
    return dest;
  }

  // -------------------------------------------------------------------------------------------------------------- //
  public    Vector2      worldToClient( Vector2 source                                      ) { return this.worldToClient( new Vector2(), source               );      }  
  public    Vector2      worldToClient( Vector2 source,               int depth             ) { return this.worldToClient( new Vector2(), source, depth        );      }  
  public    Vector2      worldToClient( Vector2 source,                          byte flags ) { return this.worldToClient( new Vector2(), source,        flags );      }
  public    Vector2      worldToClient( Vector2 source,               int depth, byte flags ) { return this.worldToClient( new Vector2(), source, depth, flags );      }
  public    Vector2      worldToClient( Vector2 dest, Vector2 source                        ) { return this.worldToClient(         dest, source,         UITRANS_ALL ); }
  public    Vector2      worldToClient( Vector2 dest, Vector2 source, int depth             ) { return this.worldToClient(         dest, source,  depth, UITRANS_ALL ); }  
  public    Vector2      worldToClient( Vector2 dest, Vector2 source,            byte flags ) { return this.worldToClient(         dest, source,   9999, flags );      }
  public    Vector2      worldToClient( Vector2 dest, Vector2 source, int depth, byte flags )
  {
    if( depth < 0 ) return source;
    if( parentTransform == null ) this.worldToClientImpl( dest, source, flags );
    else
    {
      this.worldToClientImpl( dest, source, flags );
      parentTransform.worldToClient( dest, dest.get(), depth-1, flags );
    }
    return dest;
  }
  
  // -------------------------------------------------------------------------------------------------------------- //
  public    void         setParentTransform ( UITransform parent ) { parentTransform = parent; }
  public    UITransform  getParentTransform ( )                    { return parentTransform; }
  
  // -------------------------------------------------------------------------------------------------------------- //
  protected void drawShadow( UIDrawer d )
  {
    d.boxType( UIDrawer.SHADOW );
    
    if( this.props.shadow )
      d.box( 0, 0, size.x, size.y );
  }
  // -------------------------------------------------------------------------------------------------------------- //
  protected void drawBox( UIDrawer d ) {
    d.boxType( UIDrawer.BOX );
    d.box( 0, 0, size.x, size.y );
  }
  // -------------------------------------------------------------------------------------------------------------- //
  protected void updateMomentum( )
  {
    if( !props.momentumEnabled )
      return;
  
    if( !getApp().mousePressed )
    {
      velocity.muleq( momentumDecay );
      getPos().inc( velocity );
    }     
  }
  // -------------------------------------------------------------------------------------------------------------- //
  public void draw( UIDrawer d )
  {
    transform  ( d );
    drawShadow ( d );
    drawBox    ( d );
  }


  
  // -------------------------------------------------------------------------------------------------------------- //
  public void move( Vector2 velocity )
  {
    getPos().inc( velocity );
  
    float fLen = velocity.len();
    if( fLen > 10.0f )
    {
      velocity.diveq( fLen );
      velocity.muleq( 10.0f );
      velocity.set( velocity );
    }
  }
  


  protected boolean mouseMoved()
  {
    if( props.resize || props.scale || props.rotation )
    {
      // Check the distance to an edge
      Vector2 size        = getSize();
      Vector2 topLeft     = new Vector2();
      Vector2 topRight    = topLeft.add( size.x, 0 );
      Vector2 bottomRight = topLeft.add( size );
      Vector2 bottomLeft  = topLeft.add( 0, size.y );
      
      LineSeg2 topSide      = new LineSeg2( topLeft,     topRight    );
      LineSeg2 rightSide    = new LineSeg2( topRight,    bottomRight );
      LineSeg2 bottomSide   = new LineSeg2( bottomRight, bottomLeft  );
      LineSeg2 leftSide     = new LineSeg2( bottomLeft,  topLeft     );    
      
      Vector2 clientMouse   = getClientMouse();
      closestTop    = topSide.project( clientMouse );  
      closestRight  = rightSide.project( clientMouse );  
      closestBottom = bottomSide.project( clientMouse );  
      closestLeft   = leftSide.project( clientMouse );  
      
      float radius = clientToWorld( new Vector2(2,0), UITRANS_SCALE ).x;
      
      hitTop    = closestTop.sub( clientMouse ).lenlen()    < radius*radius;
      hitBottom = closestBottom.sub( clientMouse ).lenlen() < radius*radius;
      hitLeft   = closestLeft.sub( clientMouse ).lenlen()   < radius*radius;
      hitRight  = closestRight.sub( clientMouse ).lenlen()  < radius*radius;
             
      
      
      /*if( props.rotation )
      {
        if( hitTop  || hitBottom 
         || hitLeft || hitRight  )  getCursor().set( UICursor.MOVE );
        else                        getCursor().set( UICursor.ARROW );
      }
      else
      {
             if( hitTop && hitLeft )     getCursor().set( UICursor.DragDiagonalB );
        else if( hitTop && hitRight )    getCursor().set( UICursor.DragDiagonalA );
        else if( hitTop )                getCursor().set( UICursor.DragUpDown );
        else if( hitBottom && hitRight ) getCursor().set( UICursor.DragDiagonalB );
        else if( hitRight )              getCursor().set( UICursor.DragLeftRight );
        else if( hitBottom && hitLeft )  getCursor().set( UICursor.DragDiagonalA );
        else if( hitBottom )             getCursor().set( UICursor.DragUpDown );
        else if( hitLeft )               getCursor().set( UICursor.DragLeftRight );
        else                             getCursor().set( UICursor.ARROW );
      }*/
    }
    
    // Return true if its in our client area
    return ( getClientMouseOnly() != null );
  }
  
  // -------------------------------------------------------------------------------------------------------------- //
 /* protected boolean mousePressed() {
    if( getClientMouseOnly() != null )
    {
      setFocus( true );
      return true;
    }
    return false;
  }*/
  
  // -------------------------------------------------------------------------------------------------------------- //
  protected boolean mouseDragged( ) { 
    if( hasFocus() )
    {    
      Vector2 worldMouseMove = getContainer().getView().getWorldMouseMove();
      Vector2 worldMouse     = getContainer().getView().getWorldMouse();

      if( props.resize || props.scale || props.rotation )
      {
             if( hitTop && hitLeft )     moveDragPoint ( DRAG_TOPLEFT     , worldMouse );
        else if( hitTop && hitRight )    moveDragPoint ( DRAG_TOPRIGHT    , worldMouse );
        else if( hitTop )                moveDragPoint ( DRAG_TOP         , worldMouse );
        else if( hitBottom && hitRight ) moveDragPoint ( DRAG_BOTTOMRIGHT , worldMouse );
        else if( hitRight )              moveDragPoint ( DRAG_RIGHT       , worldMouse );
        else if( hitBottom && hitLeft )  moveDragPoint ( DRAG_BOTTOMLEFT  , worldMouse );
        else if( hitBottom )             moveDragPoint ( DRAG_BOTTOM      , worldMouse );
        else if( hitLeft )               moveDragPoint ( DRAG_LEFT        , worldMouse );
      }
      
      if( props.draggable ) move ( worldMouseMove );

      return true;
    }

    return false;
  }
  
  // -------------------------------------------------------------------------------------------------------------- //
  public    String       getTypeName( )  { return "UIBox"; }
  
  public void moveDragPoint( int index, Vector2 pt )
  {
    if( !props.resize && !props.scale && !props.rotation ) return;
  
    boolean bScale  = false;
    boolean bRotate = false;
    boolean bSize   = false;

    Vector2 center = pos.add( size.mul(0.5f) );
      
    if( this.props.scale )
    {
      if( index == DRAG_TOPLEFT    || index == DRAG_TOPRIGHT    ||
          index == DRAG_BOTTOMLEFT || index == DRAG_BOTTOMRIGHT )
      {
        scale = 2.f * pt.sub( center ).len() / getSize().len();
      }
      /*
      else
      {
        if     ( index == DRAG_TOP    ) scale = pt.sub( center ).len() / closestTop.sub( center ).len();
        else if( index == DRAG_RIGHT  ) scale = pt.sub( center ).len() / closestRight.sub( center ).len(); 
        else if( index == DRAG_BOTTOM ) scale = pt.sub( center ).len() / closestBottom.sub( center ).len(); 
        else if( index == DRAG_LEFT   ) scale = pt.sub( center ).len() / closestLeft.sub( center ).len();  
      }*/
      
      bScale = true;
    }
    else if( this.props.resize )
    {     
      if( index == DRAG_TOPLEFT )       // Top Left
      {
        getSize().set( getPos().add( getSize() ).sub( pt ) );
        setPos( pt );
      }
      else if( index == DRAG_TOPRIGHT )  // Top Right
      {
        getSize().x = pt.x - getPos().x;
        getSize().y = getPos().y + getSize().y - pt.y;
        getPos().y  = pt.y;      
      }
      else if( index == DRAG_BOTTOMLEFT )  // Bottom Left
      {
        getSize().x = getPos().x + getSize().x - pt.x;
        getSize().y = pt.sub( getPos() ).y;
        getPos().x  = pt.x;
      }
      else if( index == DRAG_BOTTOMRIGHT )  // Bottom Right
      {
        getSize().set( pt.sub( getPos() ) );
      }
      else if( index == DRAG_TOP )  // Top
      {
        getSize().y = getPos().y + getSize().y - pt.y;
        getPos().y  = pt.y;      
      }
      else if( index == DRAG_RIGHT )  // Right
      {
        getSize().x = pt.x - getPos().x;
      }
      else if( index == DRAG_BOTTOM )  // Bottom
      {
        getSize().y = pt.y - getPos().y;
      }
      else if( index == DRAG_LEFT ) // Left
      {
        getSize().x = getPos().x + getSize().x - pt.x;
        getPos().x  = pt.x;      
      }
      
      bSize = true;
    }
    
    if( this.props.rotation )
    {
      // Calculate the mouse's starting location for the rotation for the edge cases
      float intitalRotation = 0;
      if     ( index == DRAG_TOP    ) intitalRotation = (float)Math.atan2( closestTop.y    - center.y , closestTop.x - center.x );
      else if( index == DRAG_RIGHT  ) intitalRotation = (float)Math.atan2( closestRight.y  - center.y , closestRight.x - center.x );
      else if( index == DRAG_BOTTOM ) intitalRotation = (float)Math.atan2( closestBottom.y - center.y , closestBottom.x - center.x );
      else if( index == DRAG_LEFT   ) intitalRotation = (float)Math.atan2( closestLeft.y   - center.y , closestLeft.x - center.x );

      // Calculate the mouse's current rotation, then offset by the correct amount
                                           rotation  = (float)Math.atan2( pt.y - center.y , pt.x - center.x );    
      if     ( index == DRAG_TOPLEFT     ) rotation += (float)Math.atan2( getSize().y *  0.5f, -getSize().x * 0.5f );
      else if( index == DRAG_TOPRIGHT    ) rotation += (float)Math.atan2( getSize().y *  0.5f,  getSize().x * 0.5f );
      else if( index == DRAG_BOTTOMLEFT  ) rotation += (float)Math.atan2( getSize().y * -0.5f, -getSize().x * 0.5f );
      else if( index == DRAG_BOTTOMRIGHT ) rotation += (float)Math.atan2( getSize().y * -0.5f,  getSize().x * 0.5f );
      else rotation -= intitalRotation;

      
      bRotate = true;
    }
    
    if( bScale  ) onScale();
    if( bRotate ) onRotate();
    if( bSize   ) { validateSize(); onResize(); }
  }
}
