/*
  Caffeine Processing Libraries
  
  (c) copyright
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
 */
 
package com.fieldfx.ui;

import com.fieldfx.math.Vector2;

// -------------------------------------------------------------------------------------------------------------- //
/** This interface defines the contract that any transformer of elements must implement to preserve the
  * child elements abilities to process events in thier own space. See {@UIView} for further details.
  * @author Jack Kern
  * @version 1.0 */
public interface UITransform {

  /** Flag used for {@link clientToWorld} and {@link worldToClient} to represent that the
      translation portion of the transformation should be applied. */
  public static final byte UITRANS_TRANSLATE  = 1;
  /** Flag used for {@link clientToWorld} and {@link worldToClient} to represent that the
      rotation portion of the transformation should be applied. */
  public static final byte UITRANS_ROTATE     = 2;
  /** Flag used for {@link clientToWorld} and {@link worldToClient} to represent that the
      scale portion of the transformation should be applied. */
  public static final byte UITRANS_SCALE      = 4;
  /** Flag used for {@link clientToWorld} and {@link worldToClient} to represent that the
      translation, rotation and scale portions of the transformation should be applied. */
  public static final byte UITRANS_ALL        = UITRANS_TRANSLATE | UITRANS_ROTATE | UITRANS_SCALE;

// -------------------------------------------------------------------------------------------------------------- //
/** Applies the transform to the current matrix. */
  public void        transform          ( UIDrawer d );
// -------------------------------------------------------------------------------------------------------------- //
/** Applies the inverse transform to the current matrix */
  public void        transformInv       ( UIDrawer d );

// -------------------------------------------------------------------------------------------------------------- //
/** Transforms a {@link Vector2} in the client coordinates of this transform's space into world coordinates outside the transform.
  * @param   dest     The {@link Vector2} object used to store the result.
  * @param   source   The {@link Vector2} containing the coordinates to be transformed.
  * @param   flags    The transform flags for this operation. Can be {@link UITRANS_TRANSLATE}, {@link UITRANS_ROTATE}, {@link UITRANS_SCALE} or {@link UITRANS_ALL}
  * @return           Returns a {@link: Vector2} containing the transformed coordinates */
  public Vector2     clientToWorld      ( Vector2 dest, Vector2 source,            byte flags );

// -------------------------------------------------------------------------------------------------------------- //
/** Transforms a {@link Vector2} in the client coordinates of this transform's space into world coordinates outside the transform.
  * @param   dest     The {@link Vector2} object used to store the result.
  * @param   source   The {@link Vector2} containing the coordinates to be transformed.
  * @param   depth    How much of the transformation stack should be applied.
  * @param   flags    The transform flags for this operation. Can be {@link UITRANS_TRANSLATE}, {@link UITRANS_ROTATE}, {@link UITRANS_SCALE} or {@link UITRANS_ALL}
  * @return           Returns a {@link: Vector2} containing the transformed coordinates */
  public Vector2     clientToWorld      ( Vector2 dest, Vector2 source, int depth, byte flags );
  
// -------------------------------------------------------------------------------------------------------------- //
/** Transforms a {@link Vector2} in world coordinates into client coordinates of this transform.
  * @param   dest     The {@link Vector2} object used to store the result.
  * @param   source   The {@link Vector2} containing the coordinates to be transformed.
  * @param   flags    The transform flags for this operation. Can be {@link UITRANS_TRANSLATE}, {@link UITRANS_ROTATE}, {@link UITRANS_SCALE} or {@link UITRANS_ALL}
  * @return           Returns a {@link: Vector2} containing the transformed coordinates */
  public Vector2     worldToClient      ( Vector2 dest, Vector2 source,            byte flags );
// -------------------------------------------------------------------------------------------------------------- //
/** Transforms a {@link Vector2} in world coordinates into client coordinates of this transform.
  * @param   dest     The {@link Vector2} object used to store the result.
  * @param   source   The {@link Vector2} containing the coordinates to be transformed.
  * @param   depth    How much of the transformation stack should be applied.
  * @param   flags    The transform flags for this operation. Can be {@link UITRANS_TRANSLATE}, {@link UITRANS_ROTATE}, {@link UITRANS_SCALE} or {@link UITRANS_ALL}
  * @return           Returns a {@link: Vector2} containing the transformed coordinates */
  public Vector2     worldToClient      ( Vector2 dest, Vector2 source, int depth, byte flags );
  
// -------------------------------------------------------------------------------------------------------------- //
/** Sets the parent transform of this transform object.
  * @param   The parent transform to set for this transform object. */
  public void        setParentTransform ( UITransform parent );

// -------------------------------------------------------------------------------------------------------------- //
/** Returns the parent transform of this transform object.
  * @return   The parent transform of this transform object.*/
  public UITransform getParentTransform ( );
    
}
