package com.fieldfx.util.serialize;

import processing.core.PApplet;
import processing.xml.XMLElement;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.List;


public class XMLSerializer implements Serializer {
  
  private PApplet                       parent    = null;
  private XMLElement                    xmlReader = null;
  private PrintWriter                   xmlWriter = null;
  private int                           depth     = 0;
  private HashMap<String, Serializable> types     = new HashMap<String, Serializable>();
  
  // ------------------------------------------------------------ //
  public XMLSerializer( PApplet parent ) {
    this.parent = parent;
  }
  
  // ------------------------------------------------------------ //
  public void load ( String fileName, Serializable object ) {
    xmlReader = new XMLElement( parent, fileName );
    serialize( "root", object, true );
    xmlReader = null;
  }
  
  // ------------------------------------------------------------ //
  public void save ( String fileName, Serializable object ) {
    xmlWriter = parent.createWriter( fileName );
    serialize( "root", object );
    xmlWriter.flush();
    xmlWriter.close();
    xmlWriter = null;
  }
  
  // ------------------------------------------------------------ //  
  public void registerType ( Serializable type ) {
    types.put( type.getType(), type );
  }
  
  // ------------------------------------------------------------ //
  public float serialize ( String name, float value ) {
    return serialize( name, value, 0.f );
  }
  
  public float serialize ( String name, float value, float defaultValue ) {
    if( isLoading() ) {
      XMLElement child = getChild( "float", name );
      if( child != null ) value = child.getFloat( "value", defaultValue );
    } else {
      xmlWriter.println( getXmlTag( "float", name, "" + value ) );
    }
    return value;
  }
  
  // ------------------------------------------------------------ //
  public int serialize ( String name, int value ) {
    return serialize( name, value, 0 );
  }
  
  public int serialize ( String name, int value, int defaultValue ) {
    if( isLoading() ) {
      XMLElement child = getChild( "int", name );
      if( child != null ) value = child.getInt( "value", defaultValue );
    } else {
      xmlWriter.println( getXmlTag( "int", name, "" + value ) );
    }
    return value;
  }
  
  public byte serialize ( String name, byte value ) {
    return (byte)serialize( name, (byte)value, 0 );
  }
	  
  
  // ------------------------------------------------------------ //
  public boolean serialize ( String name, boolean value ) {
    if( isLoading() ) {
      
      XMLElement child = getChild( "boolean", name );
      if( child != null ) {
        String _value = child.getString( "value" );
        _value = _value.trim();
        _value = _value.toLowerCase();
        
        if( _value != null && _value.equals( "true" ) ) value = true;
        else value = false;
      }
    } else {
      xmlWriter.println( getXmlTag( "boolean", name, value ? "true" : "false" ) );
    }
    return value;
  }
  
  // ------------------------------------------------------------ //
  public String serialize ( String name, String value ) {
    if( isLoading() ) {
      
      XMLElement child = getChild( "string", name );
      if( child != null ) value = child.getString( "value" );
    } else {
      xmlWriter.println( getXmlTag( "string", name, value ) );
    }
    return value;
  }
  
  // ------------------------------------------------------------ //
  // public Vector2 serialize( String name, Vector2 value ) {
  // if( isLoading() ) {
  // XMLElement child  = getChild( "Vector2", name );
  // if       ( child != null ) {
  // String   _value  =  child.getString( "value" );
  // _value  = _value.substring( 1, _value.length()-1 );
  // String[] _values = _value.split(",");
  
  // value.set( Float.parseFloat( _values[0].trim() )
  // , Float.parseFloat( _values[1].trim() ) );
  // }
  // } else {
  // xmlWriter.println( getXmlTag( "Vector2", name, value.toString() ) );
  // }
  // return value;
  // }
  
  // ------------------------------------------------------------ //
  public void serialize ( String name, Serializable object ) {
    serialize( name, object, false );
  }
  
  protected void serialize ( String name, Serializable object, boolean root ) {
    
    if( isLoading() ) {
      // First if this is the root node, then xmlReader is already set to the correct node, but let's just 
      //  do some sanity checking anyways to make sure everything goes smoothly.
      boolean actuallyRoot = root && xmlReader != null && xmlReader.getName().equals( "root" );
      XMLElement child = actuallyRoot ? xmlReader : getChild( object.getType(), name );
      
      // Now we should have the correct child element, but let's make sure something didn't go wrong.
      if( child != null ) {
        
        // Now we need to store the active node, and set it to the child node we found
        XMLElement parent = xmlReader;
        xmlReader = child;
        
        // Let the object Serialize itself
        object.serialize( this );
        
        // Finally we need to set back the active node so we can continue
        xmlReader = parent;
      }
    } else {
      if( object != null ) {
        xmlWriter.println( getXmlBeginTag( object.getType(), name ) );
        
        // Let the object Serialize itself
        object.serialize( this );
        
        xmlWriter.println( getXmlEndTag( name ) );
      }
    }
  }
  
  // ------------------------------------------------------------ //
  public <T extends Serializable> void serialize ( String name, List<T> values ) {
    if( isLoading() ) {
      values.clear();
      XMLElement child = getChild( "ArrayList", name );
      if( child != null ) {
        // Now we need to store the active node, and set it to the child node we found
        XMLElement parent = xmlReader;
        xmlReader = child;
        
        for( int index = 0; index < child.getChildCount(); ++index ) {
          XMLElement listChild = child.getChild( index );
          String typeName = listChild.getString( "type" );
          Serializable template = types.get( typeName );
          T object = (T)template.clone();
          
          serialize( listChild.getName(), object );
          values.add( object );
        }
        
        // Finally we need to set back the active node so we can continue
        xmlReader = parent;
      }
    } else {
      xmlWriter.println( getXmlBeginTag( "ArrayList", name ) );
      int index = 0;
      for( Serializable object : values )
        serialize( "child_" + ( index++ ), object );
      xmlWriter.println( getXmlEndTag( name ) );
    }
  }
  
  // ------------------------------------------------------------ //
  public boolean isLoading () {
    return xmlReader != null;
  }
  
  // ------------------------------------------------------------ //
  private XMLElement getChild ( String type, String name ) {
    // We must have an active node to read from
    if( xmlReader == null ) return null;
    
    // First try to get the child by name, and bail if we can't find it.
    XMLElement child = xmlReader.getChild( name );
    if( child == null ) return null;
    
    // Next check the type to make sure it's what we expect, if it 'aint, then bail.
    String childType = child.getStringAttribute( "type" );
    if( !childType.equals( type ) ) return null;
    
    // Finally, everything looks good so return our element
    return child;
  }
  
  // ------------------------------------------------------------ //
  private String getXmlTag ( String type, String name, String value ) {
    return getIndent() + "<" + name + " type=\"" + type + "\" value=\"" + value + "\" />";
  }
  
  private String getXmlBeginTag ( String type, String name ) {
    String beginTag = getIndent() + "<" + name + " type=\"" + type + "\">";
    ++depth;
    return beginTag;
  }
  
  private String getXmlEndTag ( String name ) {
    --depth;
    return getIndent() + "</" + name + ">";
  }
  
  private String getIndent () {
    String out = "";
    for( int indent = 0; indent < depth; ++indent )
      out += "  ";
    return out;
  }
}
