
 
package com.fieldfx.ui;

import processing.core.*;

import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;

import com.fieldfx.math.Vector2;

// -------------------------------------------------------------------------------------------------------------- //
/**
 * {@link UIElement} is the base class for all of the Caffeine Library user interface elements. <br />
 * <br />
 * {@link UIElement}s have a protected constructor because they should all be created by the
 * {@link UIFactory}. The factory will register creation with the {@link UIManager} so they are
 * automatically drawn and updated with input events.<br />
 * 
 * @author Jack W. Kern
 */
public class UIElement implements PConstants {
  
  /** This the parent UIManager that updates this element. */
  protected UIManager   parent         = null;
  
  /** The position of this element inside it's parent view. */
  protected Vector2     pos            = new Vector2(); 
  
  /** The container of the element. */
  protected UIContainer container      = null; 
  
  /** This element's creation index. Used by Serialization. */
  protected int         createIndex    = 0;
  
  /** Is this object visible or not? */
  protected boolean     visible        = true;
  
  /** Is this object a systemElement or not? */
  protected boolean     systemElement  = false;
  
  

  
// -------------------------------------------------------------------------------------------------------------- //
/** UIElement's constructor. This is a protected constructor because {@link UIElement}
 *   objects must be created by our parent's {@link UIFactory}.
 *  @param parent       The parent UIManager object that is updating this element.
 *  @param container    The parent container object that specifies the view transform 
 *                      for this object and is the next up in the chain of view 
 *                      transforms. */
  protected UIElement( UIManager parent, UIContainer container ) {
    pos.set( 0, 0 );
    this.parent    = parent;
    this.container = container;
  }  

// -------------------------------------------------------------------------------------------------------------- //
/** Returns the parent {@link UIManager} that updates this element.
 *  @return The parent {@link UIManager} class that updates this 
 *          element. */
  protected UIManager getMgr( ) { return parent; }


 protected PApplet getApp( ) { return parent.getApp();  }

// -------------------------------------------------------------------------------------------------------------- //
/** Returns the {@link UIDrawer} object that is used to render this element.
 *  @return The UIDrawer used to render this element. */ 
  protected UIDrawer getDrawer() { return parent.getDrawer(); }  

// -------------------------------------------------------------------------------------------------------------- //
/** Called to register any commands this element provides actions for, as well
 *  as implementations for those actions. Implementors of this function must follow the
 *  following pattern, because of the functions dual-nature.
<pre>
  protected boolean commands()
  {
    if     ( command("Object_Example1") ) return cmdExample1();
    else if( command("Object_Example2") ) return cmdExample2();
    else if( command("Object_Example3") ) return cmdExample3();
    
    return false;
  }
</pre>
 * @return Returns true when commands() handles a command. */
  protected boolean commands( ) { return false; }

// -------------------------------------------------------------------------------------------------------------- //
/** Called after an element is scaled. This function can be overridden to 
  * implement functionality that constrains the element, or has a need to update
  * something after a scale operation. */
  public    void    onScale         ()                     { }

// -------------------------------------------------------------------------------------------------------------- //
/** Called after an element is rotated. This function can be overridden to 
  * implement functionality that constrains the element, or has a need to update
  * something after a rotation operation. */
  public    void    onRotate        ()                     { }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Called after an element is resized. This function can be overridden to 
  *  implement functionality that constrains the element, or has a need to update
  *  something after a resize operation. */
  public    void    onResize        ()                     { }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Sets the position in the space of this element's containing view.
  * @param   x          The x coordinate of the position.
  * @param   y          The y coordinate of the position.
  * @return  Returns this element for chaining. */
  public UIElement  setPos ( float x, float y )       { pos.set(x, y);  return this;}

// -------------------------------------------------------------------------------------------------------------- //
/** Sets the position in the space of this element's containing view.
  * @param   pos        A {@link Vector2} containing the coordinates to be set. 
  *                     The position is in the space of the containing view.
  * @return  Returns this element for chaining. */
  
// -------------------------------------------------------------------------------------------------------------- //
/** Sets the position in the space of this element's containing view. 
  * @param   pos   A {@link Vector2} containing the position of the element to set. */
  public UIElement  setPos ( Vector2 pos )            { return this.setPos( pos.x, pos.y );  }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Gets the position in the space of this element's containing view.
  * @return  Returns a {@link Vector2} object containing the position of the element. */
  public Vector2    getPos( )                         { return pos;    }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Shows the element.
  * @return  Returns this element for chaining. */
  public UIElement  show( )                           { visible = true;   return this; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Hides the element
  * @return  Returns this element for chaining. */
  public UIElement  hide( )                           { visible = false;  return this; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Brings the element to the front of the z-ordered list of elements. (drawn last)
  * This does not affect a parent frames z-order which will have higher precedence.
  * @return  Returns this element for chaining. */
  public UIElement  bringToFront( )                   { parent.getView().bringToFront( this ); return this; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Sends the element class to the back of the z-ordered list of elements. (drawn first)
  * This does not affect a parent frames z-order which will have higher precedence.
  * @return  Returns this element for chaining. */
  public UIElement  sendToBack( )                     { parent.getView().sendToBack( this );   return this; }

// -------------------------------------------------------------------------------------------------------------- //
/** Returns whether this element will be drawn.
  * @return True for the element being visible, false otherwise. */
  public boolean    isVisible( )                      { return visible;  }

// -------------------------------------------------------------------------------------------------------------- //
/** Returns whether this element currently has input focus.
  * @return  True if this UIElement has focus, false otherwise. */
  public boolean   hasFocus( )                        { return getMgr().hasFocus(this);  }            
  
// -------------------------------------------------------------------------------------------------------------- //
/** Sets focus on this element.
  * @return Returns this element for chaining. */
  public UIElement setFocus( )                        { setFocus( true );                return this; }   

// -------------------------------------------------------------------------------------------------------------- //
/** Sets whether this element should have input focus or not. True gives
  * focus to the element and false removes focus from the element.
  * @param  set  True to give this element focus, false to remove focus.
  * @return Returns this element for chaining. */
  public UIElement setFocus( boolean set )           { getMgr().setFocus( this, set ); return this; }

// -------------------------------------------------------------------------------------------------------------- //
/** Called for debugging purposes.
  * This function should be overridden by all UIElements and derrived classes
  * and can be used to get the type of the element as a string.
  * @return The name of the UIElement or derrived class. */
  public  String  getTypeName( )                      { return "UIElement"; }

// -------------------------------------------------------------------------------------------------------------- //
/** Sets the parent transform for this element. 
  * @param parent   The parent {@link UITransform} to set for this element. */
  protected void setParentTransform  ( UITransform parent ) { }

// -------------------------------------------------------------------------------------------------------------- //
/** Returns the parent container for this element.
  * @return The parent {@link UIContainer} of this element. */
  protected UIContainer getContainer ( ) { return container; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Sets the parent container for this element.
  * @param container   The parent {@link UIContainer} to set for this element.
  * @return Returns this element for chaining. */
  protected UIElement setContainer ( UIContainer container ) { this.container = container; return this; }

  
// -------------------------------------------------------------------------------------------------------------- //
/** Update is called by the {@link UIManager} or parent element to allow any 
  * processing to occur in a UIElement or derrived class before drawing
  * occurs */
  protected void update ( ) { }

// -------------------------------------------------------------------------------------------------------------- //
/** Called by the {@link UIManager} or parent to draw the {@link UIElement}
  * This method draws the element. Override this method to provide custom
  * drawing for any {@link UIElement}. Drawing should use the skinning interface
  * however so users of new controls can easily skin them as they so desire.<br />
  * <br />
  * Note: Elements do not need to worry about screwing up the current transformation matrix 
  * for other elements as calls to {@link draw} are always wrapped with 
  * {@link PApplet.pushMatrix} & {@link PApplet.popMatrix}. In otherwords if you
  * are creating your own UIElement, only add {@link PApplet.pushMatrix} & 
  * {@link PApplet.popMatrix} if they are required by the drawing code you are writing. */    
  protected void  draw ( UIDrawer d )     {}

// -------------------------------------------------------------------------------------------------------------- //
/** Called by the UIManager or parent to draw the UIElement
  * This method recieves event notifications from its parent element, and translates 
  * them into function calls that can be easily overriden. 
  * @parem eventIndex   The event index of the current input event. 
  * @return True if the event was handled, false otherwise. */    
  protected boolean inputEvent ( int eventIndex )
  {
    switch( eventIndex ) {
      case KeyEvent.KEY_PRESSED:      if( this.keyPressed()    ) return true; break;
      case KeyEvent.KEY_RELEASED:     if( this.keyReleased()   ) return true; break;
      case MouseEvent.MOUSE_PRESSED:  if( this.mousePressed()  ) return true; break;
      case MouseEvent.MOUSE_RELEASED: if( this.mouseReleased() ) return true; break;
      case MouseEvent.MOUSE_CLICKED:  if( this.mouseClicked()  ) return true; break;
      case MouseEvent.MOUSE_DRAGGED:  if( this.mouseDragged()  ) return true; break;
      case MouseEvent.MOUSE_MOVED:    if( this.mouseMoved()    ) return true; break;
    }
  
    return false;
  }

// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the mousePressed event. Override this function if your element needs 
  *  to recieve mousePressed events.
  *  @return Implementers of this function should return true if the 
  *          element handles the message entirely and it doesn't want
  *          the parent element to process the event. */
  protected boolean mousePressed  ()  { return false; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the mouseReleased event. Override this function if your element needs 
 *  to recieve mouseReleased events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean mouseReleased ()  { return false; }  

// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the mouseMoved event. Override this function if your element needs 
 *  to recieve mouseMoved events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean mouseMoved    ()  { return false; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the mouseDragged event. Override this function if your element needs 
 *  to recieve mouseDragged events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean mouseDragged  () { return false; }

// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the mouseScrolled event. Override this function if your element needs 
 *  to recieve mouseScrolled events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean mouseScrolled   ( int nDelta ) { return false; }  
  
// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the mouseClicked event. Override this function if your element needs 
 *  to recieve mouseClicked events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean mouseClicked    ()             { return false; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the keyPressed event. Override this function if your element needs 
 *  to recieve keyPressed events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean keyPressed      ()             { return false; }
  
// -------------------------------------------------------------------------------------------------------------- //
/** Recieves the keyReleased event. Override this function if your element needs 
 *  to recieve keyReleased events.
 *  @return Implementers of this function should return true if the 
 *          element handles the message entirely and it doesn't want
 *          the parent element to process the event. */
  protected boolean keyReleased     ()             { return false; }  
  

}