/*
  Caffeine Processing Libraries
  
  (c) copyright
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
 */
 
package com.fieldfx.ui;

import processing.core.*;

import java.util.ArrayList;
import java.util.List;

import java.awt.event.MouseEvent;

import com.fieldfx.math.Vector2;


public class UIView implements UITransform, UIContainer {
  // ------------------------------------------------------------------------ //
  protected float                 rotation         = 0.0f;
  protected Vector2[]             pos              = new Vector2[2];
  protected Vector2               size             = new Vector2();
  protected float[]               zoom             = new float[2];
  protected boolean               moving           = false;  
  protected boolean               draggingWorld    = false;
  
  protected Vector2               preworldMouse    = null;

  protected boolean               hideChildren     = false;
  
  protected UITransform           parentTransform  = null;
  protected UIManager             parent           = null;

  
  // View Properties
  private   boolean               draggableWorld   = false;
  private   boolean               zoomableWorld    = false;
  
  // The children of this view
  private   List<UIElement>      children         = new ArrayList<UIElement>();
  
  // Mouse coordinates in this view's world space
  private   Vector2               mouseWorld       = new Vector2();
  private   Vector2               mouseWorldLast   = new Vector2();
  private   Vector2               mouseWorldMove   = new Vector2();
  
  public void enableZoom  ( )                { enableZoom( true );  }
  public void disableZoom ( )                { enableZoom( false ); }
  public void enableZoom  ( boolean enable ) { zoomableWorld  = enable; }
  
  public void enableDrag  ( )                { enableDrag(true); }
  public void disableDrag ( )                { enableDrag(false); }
  public void enableDrag  ( boolean enable ) { draggableWorld = enable; }
  
  public void hideChildren() { hideChildren = true;  }
  public void showChildren() { hideChildren = false; }
  
  public UIView getView() { 
    return this;
  }
  
  private PApplet getApp() { return parent.getApp(); }
  
  // ------------------------------------------------------------------------ //
  public UIView( UIManager parent, Vector2 size ) {
    this( parent, size.x, size.y );
  }
  // ------------------------------------------------------------------------ //
  public UIView( UIManager parent, float width, float height ) {
    this.parent  = parent;

    setSize( width, height );
    
    pos[0]      = new Vector2();
    pos[1]      = new Vector2();
    zoom[1]     = 1;
    zoom[0]     = 1;
  }
  // ------------------------------------------------------------------------ //
  public float getWidth() {
    return size.x;
  }
  // ------------------------------------------------------------------------ //
  public float getHeight() { 
    return size.y;
  }
  // ------------------------------------------------------------------------ //
  public void setSize( Vector2 size ) { 
    setSize( size.x, size.y );
  }
  // ------------------------------------------------------------------------ //
  public void setSize( float x, float y ) {
    size.set( x, y );
  }
  // ------------------------------------------------------------------------ //
  public void setPos( Vector2 pos ) {
    setPos( pos.x, pos.y );
  }
  // ------------------------------------------------------------------------ //
  public void setPos( float x, float y ) {
    pos[1].x = pos[0].x = x;
    pos[1].y = pos[0].y = y;    
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getPos() { 
    return pos[0];
  }
  public Vector2 getSize() { 
    return size; 
  }
  public void setRotation( float rot ) {
    rotation = rot;
  }
  public float getRotation() {
    return rotation;
  }
  public float getZoom() {
    return zoom[0];
  }
  public void setZoom( float zoom ) {
    this.zoom[0] = this.zoom[1] = zoom; 
  }
  public void incZoom( float zoom )
  {
    this.zoom[1] += zoom;
    this.zoom[1] = PApplet.constrain( this.zoom[1], 0.00001f, 1000.0f);
  }
  // ------------------------------------------------------------------------ //
  public void transform( UIDrawer d ) { 
    transform( d, false );
  }
  // ------------------------------------------------------------------------ //
  public void transform( UIDrawer d, boolean bVector ) {
    d.translate (  size.x *  0.5f, size.y *  0.5f );      
    d.rotate    (  rotation );
    d.scale     (  zoom[0]  );
    d.translate (  size.x * -0.5f, size.y * -0.5f );
    
    if( !bVector )
      d.translate (  pos[0].x,       pos[0].y       );    
  }
  // ------------------------------------------------------------------------ //
  public void transformInv( UIDrawer d ) { 
    transformInv( d, false );
  }
  // ------------------------------------------------------------------------ //
  public void transformInv( UIDrawer d, boolean vector )
  {
    if( !vector )
      d.translate ( -pos[0].x, -pos[0].y );
    
    d.translate (  size.x *  0.5f,  size.y *  0.5f );
    d.scale     (  1.f / zoom[0] );
    d.rotate    ( -rotation );
    d.translate (  size.x * -0.5f,  size.y * -0.5f );  
  }
  // ------------------------------------------------------------------------ //
  public Vector2 clientToWorld( Vector2 dest, Vector2 source ) { 
    return clientToWorld( dest, source, 9999, UITRANS_ALL   );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 clientToWorld( Vector2 dest, Vector2 source, int depth ) {
    return clientToWorld( dest, source, 9999, UITRANS_ALL   );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 clientToWorld( Vector2 dest, Vector2 source, byte flags ) {
    return clientToWorld( dest, source, 9999, flags);
  }
  // ------------------------------------------------------------------------ //
  public Vector2 clientToWorld( Vector2 dest, Vector2 source, int depth, byte flags ) {
    if( depth < 0 ) return source;
    if( parentTransform == null ) this.clientToWorldImpl( dest, source, flags );
    else
    {
      parentTransform.clientToWorld( dest, source, depth-1, flags );
      this.clientToWorldImpl( dest, dest.get(), flags );
    }
    return dest;
  }
  // ------------------------------------------------------------------------ //
  public Vector2 worldToClient( Vector2 dest, Vector2 source ) { 
    return worldToClient( dest, source, 9999, UITRANS_ALL );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 worldToClient( Vector2 dest, Vector2 source, int depth ) { 
    return worldToClient( dest, source, 9999, UITRANS_ALL );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 worldToClient( Vector2 dest, Vector2 source, byte flags ) { 
    return worldToClient( dest, source, 9999, flags );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 worldToClient( Vector2 dest, Vector2 source, int depth, byte flags ) {
    if( depth < 0 ) return source;
    if( parentTransform == null ) this.worldToClientImpl( dest, source, flags );
    else
    {
      this.worldToClientImpl( dest, source, flags );
      parentTransform.worldToClient( dest, dest.get(), depth-1, flags );
    }
    return dest;
  }
  // ------------------------------------------------------------------------ //
  public void setParentTransform( UITransform parent ) { 
    parentTransform = parent;
  }
  public UITransform  getParentTransform( ) { 
    return parentTransform;
  }
  // ------------------------------------------------------------------------ //
  public void moveTo( float x, float y ) {
    pos[1].x = x;
    pos[1].y = y;    
  }
  // ------------------------------------------------------------------------ //
  public void move( float dx, float dy ) {
    pos[1].x = pos[0].x + dx / zoom[0];
    pos[1].y = pos[0].y + dy / zoom[0];      
  }
  // ------------------------------------------------------------------------ //
  public void update() {
  
    float zoomSharpness = 0.25f;
    float moveSharpness = 0.5f;
    
    zoom[0] = PApplet.lerp( zoom[0], zoom[1], zoomSharpness );
    pos[0].lerp( pos[0], pos[1], moveSharpness );
      
    for( UIElement elm : children ) {
      if( elm == null                         ) continue;
      if( hideChildren && !elm.systemElement  ) continue;
      
      elm.update();
    }    
  }
  // ------------------------------------------------------------------------ //
  public void draw( UIDrawer d ) {
    // Trivial drawing code for now, just draw everything
    // Here we loop in 
    for( int i = children.size() - 1; i >= 0; --i ) {
      UIElement elm = (UIElement)children.get(i);
      
      if( elm == null || !elm.isVisible() ) continue;
      if( hideChildren && !elm.systemElement  ) continue;
      
      d.pushMatrix();
        if( !elm.systemElement )
          transform( d ); // Apply view transform
          
        elm.draw( d );
      d.popMatrix();
    }
  }
  // ------------------------------------------------------------------------ //
  /// Brings the specified UIElement or derrived class to the front of the z-ordered list of elements. (drawn last)
  /// This does not affect a parent frames z-order which will have higher precedence. @param elm The element to bring to the front.
  public UIElement sendToBack( UIElement elm ) {
    int index = children.indexOf(elm);
    if( index >=0 ) {
      children.remove( index );
      children.add( elm );
    }
    return elm;
  }
  // ------------------------------------------------------------------------ //
  /// Sends the UIElement or derrived class to the back of the z-ordered list of elements. (drawn first)
  /// This does not affect a parent frames z-order which will have higher precedence. @param elm The element to send to the back.
  public UIElement bringToFront( UIElement elm ) {
    int index = children.indexOf(elm);
    if( index >=0 ) {
      children.remove( index );
      children.add( 0, elm );
    }
    return elm;
  }
  // ------------------------------------------------------------------------ //
  public UIElement addChild( UIElement elm ) {
    UIContainer prevContainer = elm.getContainer();
    
    prevContainer.removeChild ( elm );
    children.add              ( elm );
    elm.setContainer          ( this );
    elm.setParentTransform    ( this );

    return elm;
  }
  // ------------------------------------------------------------------------ //
  public boolean removeChild( UIElement elm )
  {
    int index = children.indexOf( elm );
    if( index >= 0 )
    {
      elm.setParentTransform( null );
      children.remove( index );
      return true;
    }
    return false;
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getScreenMouseLast() {
    return new Vector2( getApp().pmouseX, getApp().pmouseY );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getScreenMouse() {
    return new Vector2( getApp().mouseX, getApp().mouseY );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getScreenMouseMoved() {
    return getScreenMouse().sub( getScreenMouseLast() );
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getWorldMouseLast() {
    Vector2 vParent = getScreenMouseLast();
    clientToWorld( mouseWorldLast, vParent );  
    return mouseWorldLast; 
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getWorldMouse()
  {
    Vector2 vParent = getScreenMouse();
    clientToWorld( mouseWorld, vParent );  
    return mouseWorld;  
  }
  // ------------------------------------------------------------------------ //
  public Vector2 getWorldMouseMove()
  {
    mouseWorldMove.set( getWorldMouse() );    
    Vector2 vParent = getScreenMouseLast();
    Vector2 vOut    = new Vector2();
    clientToWorld( vOut, vParent );
    mouseWorldMove.dec( vOut );      
    return mouseWorldMove;
  }
  //---------------------------------------------------------------------------------//
  protected void mouseScrolled( int nDelta ) {
    for( UIElement elm : children ) {
      if( elm == null ) continue;
      if( elm.mouseScrolled(nDelta) )
        return;
    }
    if( zoomableWorld ) {
      float zoom = -0.08f * nDelta * this.zoom[0];
      incZoom( zoom );
    }
  }
  // ------------------------------------------------------------------------ //
  protected boolean isMainView() { return this == parent.getView(); }
  //---------------------------------------------------------------------------------//
  protected boolean inputEvent( int eventIndex ) {
    if( isMainView() ) {
      if( parent.hasFocus() ) {
        return parent.getFocus().inputEvent( eventIndex );
      }
    }

    if( eventIndex == MouseEvent.MOUSE_PRESSED )
    for( UIElement elm : children ) {
      if( elm instanceof UIBox ) {
        UIBox box = (UIBox)elm;
        if( box.getClientMouseOnly() != null ) {
          parent.setFocus( box, true );
          return true;
        }
      }
    }
    
    
    if( isMainView() ) {
      switch( eventIndex ) {
        case MouseEvent.MOUSE_PRESSED: return viewDragInit();
        case MouseEvent.MOUSE_DRAGGED: return updateViewDrag();
      }
    
      if( eventIndex == MouseEvent.MOUSE_PRESSED ) {
          parent.clearFocus();
      }
    }
    
    return false;
  }
  // ------------------------------------------------------------------------ //
  private boolean mouseEvent( int eventIndex ) {
    return eventIndex == MouseEvent.MOUSE_PRESSED ||
           eventIndex == MouseEvent.MOUSE_CLICKED ||
           eventIndex == MouseEvent.MOUSE_DRAGGED ||
           eventIndex == MouseEvent.MOUSE_MOVED;
  }
  //---------------------------------------------------------------------------------//
  private boolean viewDragInit() {
    // Make sure dragging of the UI world is alowed
    if( !draggableWorld ) return false;
    
    if( getApp().mouseButton == PConstants.LEFT ) {
      preworldMouse = getWorldMouse().get();
      draggingWorld = true;
      return true;
    }  
    
    return false;
  }
  //---------------------------------------------------------------------------------//
  private boolean updateViewDrag() {
    // Make sure dragging of the UI world is allowed
    if( !draggableWorld ) return false;
  
    if( getApp().mouseButton == PConstants.LEFT && draggingWorld ) {
      Vector2 wm = getWorldMouse();
      move( wm.x - preworldMouse.x, wm.y - preworldMouse.y );
      return true;
    }  
    
    return false;
  }  

  // ------------------------------------------------------------------------ //
  private void worldToClientImpl( Vector2 dest, Vector2 source, byte flags ) {
    dest.set ( source.x,      source.y      );
    
    if( (flags & UITRANS_TRANSLATE) > 0 )
    {
      dest.inc   ( pos[0].x,      pos[0].y      );
      dest.dec   ( size.x * 0.5f, size.y * 0.5f );
    }
    
    if( (flags & UITRANS_SCALE) > 0 )
      dest.muleq ( zoom[0]  );

    if( (flags & UITRANS_ROTATE) > 0 )
      dest.rot   ( rotation );
    
    if( (flags & UITRANS_TRANSLATE) > 0 )
    {    
      dest.inc   ( size.x * 0.5f, size.y * 0.5f );
    }
  }
  
  // ------------------------------------------------------------------------ //
  private void clientToWorldImpl( Vector2 dest, Vector2 source, byte flags ) {
    dest.set( source.x, source.y );
    
    if( (flags & UITRANS_TRANSLATE) > 0 )
      dest.dec( size.x * 0.5f, size.y * 0.5f );
    
    if( (flags & UITRANS_ROTATE) > 0 )
      dest.rot( -rotation );
    
    if( (flags & UITRANS_SCALE) > 0 )
      dest.muleq( 1.f / zoom[0] );
    
    
    if( (flags & UITRANS_TRANSLATE) > 0 ) {
      dest.inc   ( size.x * 0.5f, size.y * 0.5f ); 
      dest.dec   ( pos[0].x,      pos[0].y      );
    }
  }
}
